
#include "sgrinder.h"

int SGrinder::grind(AcptSector * s, AcptCoord ** clist, int numsegs)
{

   int i;
   AcptCoord2H * cplist;
   AcptCoord * cp2list;

   if (numsegs == 0)
      NSegs = DEFAULT_NSEGS;
   else
      NSegs = numsegs;

   NumPts = 0;
   if (acptIsSectorH(s) || acptIsSectorCPArc(s))
   {
      NumCtrlPts = s->u.SectorH.NCoords;
      cplist = s->u.SectorH.coords;
      cPts = new coord2h[NumCtrlPts];
      for(i = 0; i < NumCtrlPts; i++)
      {
         cPts[i] = coord2h(cplist[i]);
      }

      if (acptIsSectorH(s))
      {
         if (NumCtrlPts > 3)
            grind_bspline();
         else
            grind_bezier();
      }
      else
         grind_arc();

      delete cPts;
   }
   else
   {
      if (acptIsCircle(s))
      {
         cp2list = s->u.SimpleSector.coords;
         c2Pts = new coord2[2];
         for(i = 0; i < 2; i++)
         {
            c2Pts[i] = coord2(cp2list[i]);
         }
         grind_circle();
         delete c2Pts;
      }
      else
      {
         NumPts = s->u.SimpleSector.NCoords;
         cp2list = s->u.SimpleSector.coords;
         gPts = new AcptCoord[NumPts];
         for(i = 0; i < NumPts; i++)
         {
            gPts[i].x = cp2list[i].x;
            gPts[i].y = cp2list[i].y;
         }
      }
   }

   if (NumPts > 0)
      *clist = gPts;

   return NumPts;

}

void SGrinder::grind_circle()
{
   // Grind circle using greedy trigonometric algorithm

   coord2 cc, c0, cr;
   double r, t0, t1, t, theta;
   double pi = acos(-1.0);
   int i;

   // Get control points

   cc = c2Pts[0];   // center point
   c0 = c2Pts[1];   // radius

   // Get circle parameters and start/end angles

   cr = c0 - cc;
   r = sqrt(cr.x * cr.x + cr.y * cr.y);
   t0 = 0.0;
   t1 = asin(1.0) * 4.0;    // 2 * pi
   t = t1 - t0;

   // Generate points

   NumPts = (int) ceil(2.0 * (double) NSegs * fabs(t) / pi);
   gPts = new AcptCoord[NumPts];
   for(i = 0; i < NumPts; i++)
   {
      theta = t0 + (i * t / NumPts);
      gPts[i].x = (r * cos(theta)) + cc.x;
      gPts[i].y = (r * sin(theta)) + cc.y;

   }

}

void SGrinder::grind_arc()
{

   // Grind elliptical arc using greedy trigonometric algorithm

   coord2 c0, cc, c1, p1, p2;
   coord2 cd1, cd2;
   double A, B, sin_t, cos_t;
   double ta, tb, tc, td;
   double r, r1, r2, t0, t1, t;
   double p1x1, p1y1, p2x1, p2y1;
   double theta, x0, y0;
   double pi = acos(-1.0);
   bool ccw;
   int NumSegs, i;

   // Get control points

   c0 = cPts[0].as_coord2();
   cc = cPts[1].as_coord2();
   c1 = cPts[2].as_coord2();
   ccw = ! arc_is_clockwise();
   p1 = c0 - cc;
   p2 = c1 - cc;

   // Get ellipse parameters

   get_ellipse_parameters(p1, p2, &A, &B, &sin_t, &cos_t);
   if (B == 0.0)
   {

      // Degenerate result

      gPts = new AcptCoord[2];
      gPts[0].x = c0.x;
      gPts[0].y = c0.y;
      gPts[1].x = c1.x;
      gPts[1].y = c1.y;
      return;
   }

   // Find affine transformation of unit circle

   ta = A * cos_t;
   tb = - B * sin_t;
   tc = A * sin_t;
   td = B * cos_t;

   // Find start and end angles

   cd1 = coord2(p1);
   cd2 = coord2(p2);
   r = ta * td - tb * tc;
   if (r == 0.0)
   {
      p1x1 = cd1.x;
      p1y1 = cd1.y;
      p2x1 = cd2.x;
      p2y1 = cd2.y;
   }
   else
   {
      p1x1 = (td * cd1.x - tb * cd1.y) / r;
      p1y1 = (ta * cd1.y - tc * cd1.x) / r;
      p2x1 = (td * cd2.x - tb * cd2.y) / r;
      p2y1 = (ta * cd2.y - tc * cd2.x) / r;
   }

   r1 = sqrt(p1x1 * p1x1 + p1y1 * p1y1);
   p1 = coord2(p1x1, p1y1) / r1;
   t0 = p1.angle_of();
   r2 = sqrt(p2x1 * p2x1 + p2y1 * p2y1);
   p2 = coord2(p2x1, p2y1) / r2;
   t1 = p2.angle_of();
   if (ccw && (t1 < t0))
      t1 = t1 + (pi * 2.0);
   else
   {
      if ((! ccw) && (t1 > t0))
         t1 = t1 - (pi * 2);
   }
   t = t1 - t0;

   // Generate points

   NumSegs = (int) ceil(2.0 * ((double) NSegs) * fabs(t) / pi);
   NumPts = NumSegs + 1;
   gPts = new AcptCoord[NumPts];
   gPts[0].x = c0.x;
   gPts[0].y = c0.y;
   for(i = 1; i < NumPts - 1; i++)
   {
      theta = t0 + (i * t / NumSegs);
      x0 = cos(theta);
      y0 = sin(theta);
      gPts[i].x = (ta * x0 + tb * y0) + cc.x;
      gPts[i].y = (tc * x0 + td * y0) + cc.y;
   }
   gPts[NumPts - 1].x = c1.x;
   gPts[NumPts - 1].y = c1.y;

}

bool SGrinder::arc_is_clockwise()
{

   coord2 c0, c1, c2;
   double x1, y1, x2, y2, d, w;
   bool cw;

   // Determine direction of control points

   c0 = cPts[0].as_coord2();
   c1 = cPts[1].as_coord2();
   c2 = cPts[2].as_coord2();

   x1 = c1.x - c0.x;
   y1 = c1.y - c0.y;
   x2 = c2.x - c0.x;
   y2 = c2.y - c0.y;
   d = x1 * y2 - x2 * y1;    // if d < 0, points are clockwise
   if (d == 0.0)
      return false;          // degenerate so it doesn't matter

   // Now compare w value

   w = cPts[1].w;
   if (d < 0)
      cw = ((w > -1) && (w < 0)) || (w > 1);
   else
      cw = (w < -1) || ((w > 0) && (w < 1));

   return cw;

}

void SGrinder::get_ellipse_parameters(coord2 p1, coord2 p2, double *A, double *B, double *sin_t, double *cos_t)
{

   // Solves ellipse based on parallel diameters

   double A2_d, B2_d, A_d, B_d;
   double sin_d, cos_d, cross, dot;
   double r1, r2, r3, sin_a, cos_a;
   double sin_2b, cos_2b, sin_b, cos_b;
   double C2, C;
   coord2 temp, p3;
   coord2 m1, v1, v2;

   // Weed out degenerate cases

   A2_d = p1.x * p1.x + p1.y * p1.y;
   B2_d = p2.x * p2.x + p2.y * p2.y;
   A_d = sqrt(A2_d);
   B_d = sqrt(B2_d);
   sin_d = p1.y / A_d;
   cos_d = p1.x / A_d;
   cross = p1.x * p2.y - p2.x * p1.y;
   dot = p1.x * p2.x + p1.y * p2.y;
   if (cross == 0.0)
      B_d = 0.0;
   if ((cross == 0.0) || (dot == 0.0) || (A_d == B_d))
   {
      *A = A_d;
      *B = B_d;
      *sin_t = sin_d;
      *cos_t = cos_d;
      return;
   }

   // Swap defining points if B > A

   if (B_d > A_d)
   {
      temp = coord2(p1);
      p1 = coord2(p2);
      p2 = coord2(temp);
   }

   // Find alpha

   p3 = p1 + p2;
   r3 = sqrt(p3.x * p3.x + p3.y * p3.y);
   sin_a = p3.y / r3;
   cos_a = p3.x / r3;

   // Find beta

   m1 = p3 / 2;
   v1 = p1 - m1;
   v2 = p3 - m1;
   r1 = sqrt(v1.x * v1.x + v1.y * v1.y);
   r2 = sqrt(v2.x * v2.x + v2.y * v2.y);

   sin_2b = (v1.x * v2.y - v2.x * v1.y) / (r1 * r2);
   cos_2b = (v1.x * v2.x + v1.y * v2.y) / (r1 * r2);
   if (fabs(cos_2b) > 1.0)
   {
      // Another degenerate case
      // (probably slips past the previous test through floating point error)

      *A = A_d;
      *B = 0.0;
      *sin_t = sin_d;
      *cos_t = cos_d;
      return;
   }

   sin_b = sqrt((1.0 - cos_2b) / 2.0);
   if (sin_2b < 0)
      sin_b = - sin_b;
   cos_b = sqrt((1.0 + cos_2b) / 2.0);
   if (cos_2b < 0)
      cos_b = - cos_b;

   // theta = alpha - beta

   *sin_t = sin_a * cos_b - sin_b * cos_a;
   *cos_t = cos_a * cos_b + sin_a * sin_b;

   // Find A and B

   C2 = A2_d + B2_d;
   C = sqrt(C2);
   *A = fabs(C * cos_b);
   *B = fabs(C * sin_b);

}

void SGrinder::grind_bezier()
{

   // Grind a rational bezier curve
   // (TO DO:  develop a more efficient algorithm for quadratic case)

   int n, iU, i;
   double x, y, denom, U;
   double bc, bp, xi, yi, wi;

   n = NumCtrlPts - 1;
   NumPts = NSegs + 1;
   gPts = new AcptCoord[NumPts];
   for(iU = 0; iU < NumPts; iU++)
   {

      x = 0.0;
      y = 0.0;
      denom = 0.0;
      U = (double) iU / (double) NSegs;
      for(i = 0; i < NumCtrlPts; i++)
      {
         bc = (double) factorial(n) / (double) factorial(n - i) / (double) factorial(i);
         bp = bc * pow(U, i) * pow((1 - U), (n - i));
         xi = cPts[i].x;
         yi = cPts[i].y;
         wi = cPts[i].w;
         x += bp * xi;               // wi is already multiplied in
         y += bp * yi;
         denom += bp * wi;
      }
      gPts[iU].x = x / denom;
      gPts[iU].y = y / denom;

   }

}

void SGrinder::grind_bspline()
{

   // Grind a uniform rational b-spline (quadratic)
   // [Based on code by Kenny Hoff]

   int i, iU;
   int Order, filler, TotSegs;
   double U;
   coord2h Pt;

   // Set Order = actual order + 1

   Order = 3;      // quadratic b-spline

   // Create normalized, clamped knot vector

   NumKnots = NumCtrlPts + Order;
   filler = NumKnots - (Order * 2);
   Knots = new double[NumKnots];
   for(i = 0; i < Order; i++)
   {
      Knots[i] = 0.0;
      Knots[Order + filler + i] = 1.0;
   }
   for(i = 0; i < filler; i++)
   {
      Knots[Order + i] = (double) (i + 1) / (double) (filler + 1);
   }

   // Do it!

   TotSegs = NSegs * (NumCtrlPts - 2);
   NumPts = TotSegs + 1;
   gPts = new AcptCoord[NumPts];
   for(iU = 0; iU < NumPts; iU++)
   {
      Pt.Set(0.0, 0.0, 0.0);
      U = (double) iU / (double) TotSegs;
      for(i = 0; i < NumCtrlPts; i++)
         Pt = Pt + (cPts[i] * R(i, Order, U));
      gPts[iU].x = Pt.x;
      gPts[iU].y = Pt.y;
   }
   delete Knots;

}

double SGrinder::R(int i, int k, double u)
{

   int j;
   double Numer, Denom;

   Numer = B(i, k, u) * cPts[i].w;
   Denom = 0.0;
   for (j = 0; j < NumCtrlPts; j++)
      Denom += ( B(j, k, u) * cPts[j].w );
   if (Denom == 0.0)
      return 0.0;

   return (Numer / Denom);

}

double SGrinder::B(int k, int m, double t)
{

   double FirstTerm, SecondTerm, Denom;

   if (m == 1)
      return ((t >= Knots[k] && t <= Knots[k + 1]) ? 1.0 : 0.0);

   Denom = Knots[k + m - 1] - Knots[k];
   if (Denom == 0.0)
      FirstTerm = 0.0;
   else
      FirstTerm = ((t - Knots[k]) / Denom) * B(k, m - 1, t);

   Denom = Knots[k + m] - Knots[k + 1];
   if (Denom == 0.0)
      SecondTerm = 0.0;
   else
      SecondTerm = ((Knots[k+m] - t) / Denom) * B(k + 1, m - 1,t);

   return (FirstTerm + SecondTerm);

}

int SGrinder::factorial(int n)
{

   if (n == 0)
      return 1;
   return (n * factorial(n - 1));

}

